@@ -1,7 +1,9 @@
 setx.default <- function(object, fn = list(numeric = mean, ordered =
                                    median, other = mode), data = NULL,
                          cond = FALSE, counter = NULL, ...){
-  mf <- match.call()
+  mc <- match.call()
+  if (class(object)[1]=="MI")
+    object <- object[[1]]
   mode <- function(x){
     tb <- tapply(x, x, length)
     if(is.factor(x))
@@ -50,19 +52,33 @@ setx.default <- function(object, fn = list(numeric = mean, ordered =
       stop("min cannot be calculated for this data type")
     return(value)
   }
-  if (any(class(object)=="vglm"))
-    tt <- object@terms$terms
+  tt <- terms(object)
+  tt.attr <- attributes(tt)
+  env <- tt.attr$.Environment
+  if (is.null(env))
+    env <- parent.frame()
+  ## original data
+  if (is.null(data))
+    if (is.data.frame(object$data))
+      dta <- object$data
+    else
+      dta <- eval(object$call$data, envir = env)
   else
-    tt <- terms(object)
-  if (is.null(data)) 
-    dta <- eval(object$call$data, sys.parent())
+    dta <- as.data.frame(data)
+  ## extract variables we need
+  mf <- model.frame(tt, data = dta, na.action = na.pass)
+  if(any(class(tt)=="multiple"))
+    vars<-unlist(c(attr(tt,"depVars"),attr(tt,"indVars")),use.names=FALSE)
   else
-    dta <- data
-  data <- dta <- dta[complete.cases(model.frame(tt, dta)), ]
-  vars <- names(dta)
+  vars <- all.vars(tt)
+  if (!is.null(tt.attr$response) && tt.attr$response)
+    resvars <- all.vars(tt.attr$variables[[1+tt.attr$response]])
+  else
+    resvars <- NULL
+  data <- dta[complete.cases(mf), names(dta)%in%vars, drop=FALSE]
   if (!is.null(counter)) {
-    if (!any(counter == names(model.frame(tt, dta))))
-      stop(paste("the variable specified for counter is not used in the model"))
+    if (!any(counter == vars))
+      stop("the variable specified for counter is not used in the model")
     treat <- data[, names(data)==counter]
     if(is.numeric(treat)) {
       data[treat==1, names(data)==counter] <- 0
@@ -76,7 +92,7 @@ setx.default <- function(object, fn = list(numeric = mean, ordered =
         data[treat==0, names(data)==counter] <- lev[2]
       }
       else
-        stop(paste("counter only takes a binary variable"))
+        stop("counter only takes a binary variable")
     }
     else if(is.logical(treat)) {
       treat <- as.numeric(treat)
@@ -84,23 +100,22 @@ setx.default <- function(object, fn = list(numeric = mean, ordered =
       data[treat==0, names(data)==counter] <- TRUE
     }
     else
-      stop(paste("not supported variable type for counter"))
+      stop("not supported variable type for counter")
     if(!cond)
-      stop(paste("if counter is specified, cond must be TRUE"))
+      stop("if counter is specified, cond must be TRUE")
   }
   if (cond) {
     if (is.null(data)) 
-      stop(paste("if cond = TRUE, you must specify the data frame."))
+      stop("if cond = TRUE, you must specify the data frame.")
+    if (is.null(mc$fn))
+      fn <- NULL
     if (!is.null(fn)) {
-      warning(paste("when cond = TRUE, fn is coerced to NULL"))
+      warning("when cond = TRUE, fn is coerced to NULL")
       fn <- NULL
     }
-    if (class(object)[1] == "vglm")
-      Y <- model.response(model.frame(tt, data=dta), "any")
-    else
-      Y <- model.extract(model.frame(tt, data=dta), "response")
+    maxl <- nrow(data)
   }
-  else  if (!is.null(fn)) {
+  else if (!is.null(fn)) {
     if (is.null(fn$numeric) || !is.function(fn$numeric)) {
       warning("fn$numeric coerced to mean().")
       fn$numeric <- mean
@@ -120,57 +135,60 @@ setx.default <- function(object, fn = list(numeric = mean, ordered =
       warning("the only available fn for other is mode.")
       fn$other <- mode
     }
-    for (i in 1:ncol(dta)) {
-      v <- na.omit(dta[,i])
-      if (any(vars[i] == names(model.frame(tt, dta)))) {
-        if (is.numeric(v))
-          value <- lapply(list(v), fn$numeric)[[1]]
-        else if (is.ordered(v)) 
-          value <- lapply(list(v), fn$ordered)[[1]]
+    for (i in 1:ncol(data)) {
+      if (!(colnames(data)[i] %in% resvars)) {
+        if (is.numeric(data[,i]))
+          value <- lapply(list(data[,i]), fn$numeric)[[1]]
+        else if (is.ordered(data[,i])) 
+          value <- lapply(list(data[,i]), fn$ordered)[[1]]
         else 
-          value <- lapply(list(v), fn$other)[[1]]
+          value <- lapply(list(data[,i]), fn$other)[[1]]
         data[,i] <- value
       }
     }
-    opt <- vars[na.omit(pmatch(names(mf), vars))]
     maxl <- 1
-    if (length(opt) > 0)
-      for (i in 1:length(opt)) {
-        value <- eval(mf[[opt[i]]], sys.parent())
-        lv <- length(value)
-        if (lv>1)
-          if (maxl==1 || maxl==lv) {
-            maxl <- lv
-            data <- data[1:lv,]
-          }
-          else
-            stop("vector inputs should have the same length.")
-        if (is.factor(data[,opt[i]]))
-          data[,opt[i]] <- list(as.factor(value))
-        else if (is.numeric(data[,opt[i]]))
-          data[,opt[i]] <- list(as.numeric(value))
-        else if (is.logical(data[,opt[i]]))
-          data[,opt[i]] <- list(as.logical(value))
+  } else {
+    maxl <- nrow(data)
+  }
+  opt <- vars[na.omit(pmatch(names(mc), vars))]
+  if (length(opt) > 0)
+    for (i in 1:length(opt)) {
+      value <- eval(mc[[opt[i]]], envir = env)
+      lv <- length(value)
+      if (lv>1)
+        if (maxl==1 || maxl==lv) {
+          maxl <- lv
+          data <- data[1:lv,]
+        }
         else
-          data[,opt[i]] <- list(value)
-      }
-    data <- data[1:maxl,]
+          stop("vector inputs should have the same length.")
+      if (is.factor(data[,opt[i]]))
+        data[,opt[i]] <- list(factor(value, levels=levels(data[,opt[i]])))
+      else if (is.numeric(data[,opt[i]]))
+        data[,opt[i]] <- list(as.numeric(value))
+      else if (is.logical(data[,opt[i]]))
+        data[,opt[i]] <- list(as.logical(value))
+      else
+        data[,opt[i]] <- list(value)
+    }
+  data <- data[1:maxl,]
+  
+  if (!is.data.frame(data)) {
+    data <- data.frame(data)
+    names(data) <- vars
   }
-  X <- model.matrix(delete.response(tt), data = data, contrasts =
-                    if (length(object$contrasts)>0) object$contrasts
-                    else NULL, xlev = object$xlevels, ...)
-  row.names(X) <- 1:nrow(X)
-  X <- as.matrix(X)
-  class(X) <- "setx"
   if (cond) {
-    X <- cbind(Y, X)
+    X <- model.frame(tt, data = dta)
     if (!is.null(counter)) {
       X <- list(treat=X[treat==1,], control=X[treat==0,])
-      class(X$treat) <- class(X$control) <- c("setx", "cond")
+      class(X$treat) <- class(X$control) <- c("data.frame", "cond")
       class(X) <- "setx.counter"
     }
     else
-      class(X) <- c("setx", "cond")
+      class(X) <- c("data.frame", "cond")
   }
+  else
+    X <- as.data.frame(model.matrix(tt, data = data))
+
   return(X)
 }
